import math
import itertools

import numpy as np

from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter
from qiskit.quantum_info import Operator
from qiskit.circuit.library import RZGate

from .constrained_mixer import Constrained

from qaoa.util import dicke_circuit


class Grover(Constrained):
    def __init__(self, k) -> None:
        self.k = k

    def create_circuit(self):
        self.U_s = dicke_circuit.DickeCircuit(self.N_qubits, self.k)#.decompose().decompose().decompose()
        self.U_s_dagger = self.U_s.inverse()

        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)

        Beta = Parameter("x_beta")
        rz = RZGate(Beta).control(self.N_qubits - 1)

        self.circuit.barrier()
        self.circuit.compose(self.U_s_dagger, q, inplace=True)
        self.circuit.barrier()
        self.circuit.x(range(self.N_qubits))
        self.circuit.barrier()
        self.circuit.append(rz, self.circuit.qubits)
        self.circuit.barrier()
        self.circuit.x(range(self.N_qubits))
        self.circuit.barrier()
        self.circuit.compose(self.U_s, self.circuit.qubits, inplace=True)
        self.circuit.barrier()

    def compute_feasible_subspace(self):
        print("Its now computing the feasible subspace")
        self.B.clear()
        for combination in itertools.combinations(range(self.N_qubits), self.k):
            current_state = ["0"] * selfN_qubits
            for index in combination:
                current_state[index] = "1"
            self.B.append("".join(current_state))

    def isFeasible(self, string):
        x = self.__str2np(string)
        constraint = np.sum(x) - self.k
        return math.isclose(constraint, 0, abs_tol=1e-7)

    #def compute_feasible_subspace(self):
    #    print("Its now computing the feasible subspace")
    #    self.B.clear()
    #    for combination in itertools.combinations(
    #        range(self.params["N_qubits"]), self.k
    #    ):
    #        current_state = ["0"] * self.params["N_qubits"]
    #        for index in combination:
    #            current_state[index] = "1"
    #        self.B.append("".join(current_state))

    #def isFeasible(self, string):
    #    x = self.__str2np(string)
    #    constraint = np.sum(x) - self.k
    #    return math.isclose(constraint, 0, abs_tol=1e-7)



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter

from .base_initialstate import InitialState


class Plus(InitialState):
    def __init__(self) -> None:
        super().__init__()

    def create_circuit(self):
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        self.circuit.h(q)



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter
from qiskit.circuit.library import XXPlusYYGate

from .constrained_mixer import Constrained

import math
import itertools

import numpy as np


class PauliString:
    def __init__(self, *args, **kwargs) -> None:
        raise NotImplementedError


class XY(Constrained):
    def __init__(self, method="chain") -> None:
        self.method = method

    def create_circuit(self):
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        # self.best_mixer_terms, self.logical_X_operators = self.__XYMixerTerms()

        Beta = Parameter("x_beta")
        scale = 0.5  # Since every logical X has two stabilizers
        for i in range(self.N_qubits - 1):
            # Hard coded XY mixer
            current_gate = XXPlusYYGate(scale * Beta)
            self.circuit.append(current_gate, [i, i + 1])
        if self.method == "ring":
            current_gate = XXPlusYYGate(scale * Beta)
            self.circuit.append(current_gate, [self.N_qubits - 1, 0])

    def compute_feasible_subspace(self):
        print("Its now computing the feasible subspace")
        self.B.clear()
        for combination in itertools.combinations(range(self.N_qubits), self.k):
            current_state = ["0"] * selfN_qubits
            for index in combination:
                current_state[index] = "1"
            self.B.append("".join(current_state))

    def isFeasible(self, string):
        x = self.__str2np(string)
        constraint = np.sum(x) - self.k
        return math.isclose(constraint, 0, abs_tol=1e-7)

    def __XYMixerTerms(self):
        logical_X_operators = [None] * (selfN_qubits - 1)
        mixer_terms = {}
        scale = 0.5  # 1/size, size of stabilizer space
        for i in range(self.params["N_qubits"] - 2):
            logical_X_operator = ["I"] * (self.params["N_qubits"] - 1)
            logical_X_operator[i] = "X"
            logical_X_operator[i + 1] = "X"
            logical_X_operator = "".join(logical_X_operator)
            logical_X_operators[i] = logical_X_operator

            mixer_terms[logical_X_operator] = [PauliString(scale, logical_X_operator)]

            YY_operator = ["I"] * (self.params["N_qubits"] - 1)
            YY_operator[i] = "Y"
            YY_operator[i + 1] = "Y"
            YY_operator = "".join(YY_operator)

            mixer_terms[logical_X_operator].append(PauliString(scale, YY_operator))

        return mixer_terms, logical_X_operators



import numpy as np


class Statistic:
    """
    See: https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf
    """

    def __init__(self, alpha=1):
        self.alpha = alpha
        self.reset()

    def reset(self):
        self.W = 0
        self.maxval = float("-inf")
        self.minval = float("inf")
        self.E = 0
        self.S = 0
        self.all_values = np.array([])

    def add_sample(self, value, weight):
        self.W += weight
        tmp_E = self.E
        self.maxval = max(value, self.maxval)
        self.minval = min(value, self.minval)
        self.E += weight / self.W * (value - self.E)
        self.S += weight * (value - tmp_E) * (value - self.E)
        if self.alpha < 1:
            idx = np.searchsorted(self.all_values, value)
            self.all_values = np.insert(
                self.all_values, idx, np.ones(int(weight)) * value
            )

    def get_E(self):
        return self.E

    def get_Variance(self):
        return self.S / (self.W - 1)

    def get_max(self):
        return self.maxval

    def get_min(self):
        return self.minval

    def get_CVaR(self):
        if self.alpha < 1:
            alphaK = int(np.round(self.alpha * len(self.all_values)))
            cvar = np.sum(self.all_values[-alphaK:]) / alphaK
            return cvar
        else:
            return self.get_E()



from qaoa.util import dicke_circuit

from .base_initialstate import InitialState


class DickeState(InitialState):
    def __init__(self, k) -> None:
        super().__init__()
        self.k = k

    def create_circuit(self):
        decompose = True
        if decompose:
            self.circuit = (
                dicke_circuit.DickeCircuit(self.N_qubits, self.k)
                .decompose()
                .decompose()
                .decompose()
            )
        else:
            self.circuit = dicke_circuit.DickeCircuit(self.N_qubits, self.k)
